@@ -1,13 +1,15 @@
 import io
+import locale
+import os
 import platform
 import subprocess
+from collections import OrderedDict
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Tuple, Type, Union
-import os 
+
 from ninja.ninja_syntax import Writer
-from collections import OrderedDict
+
 from ccimport import compat
-import locale
 
 LOCALE_TO_MSVC_DEP_PREFIX = {
     "en": "Note: including file:",
@@ -24,14 +26,22 @@ if _LOC in LOCALE_TO_MSVC_DEP_PREFIX:
 else:
     _LOC_SPLIT = _LOC.split("_")
     if _LOC_SPLIT[0] in LOCALE_TO_MSVC_DEP_PREFIX:
-        DEFAULT_MSVC_DEP_PREFIX = LOCALE_TO_MSVC_DEP_PREFIX[_LOC_SPLIT[0]]        
-
+        DEFAULT_MSVC_DEP_PREFIX = LOCALE_TO_MSVC_DEP_PREFIX[_LOC_SPLIT[0]]
 
 ALL_SUPPORTED_COMPILER = set(['cl', 'nvcc', 'g++', 'clang++'])
 ALL_SUPPORTED_LINKER = set(['cl', 'nvcc', 'g++', 'clang++'])
 
 _ALL_OVERRIDE_FLAGS = (set(["/MT", "/MD", "/LD", "/MTd", "/MDd", "/LDd"]), )
 
+ALL_SUPPORTED_PCH_COMPILER = set(['g++', 'clang++'])
+PCH_COMPILER_NEED_SOURCE = set(['cl'])
+
+COMPILER_TO_PCH_SUFFIX = {
+    "clang++": ".pch",
+    "g++": ".gch",
+}
+
+
 def _make_unique_name(unique_set, name, max_count=10000):
     if name not in unique_set:
         unique_set.add(name)
@@ -53,7 +63,6 @@ class UniqueNamePool:
         return _make_unique_name(self.unique_set, name, self.max_count)
 
 
-
 def _list_none(val):
     if val is None:
         return []
@@ -75,6 +84,7 @@ def _filter_unsupported_compiler(compilers: List[str]):
             supported.append(c)
     return supported
 
+
 def _override_flags(major_flags, minor_flags):
     """if a flag exists in _ALL_OVERRIDE_FLAGS, the one
     in major flag will override the one in minor.
@@ -93,12 +103,19 @@ def _override_flags(major_flags, minor_flags):
     return new_flags
 
 
+def _unify_path(path: Union[str, Path]):
+    path = Path(path)
+    if path.exists():
+        return path.resolve()
+    return path
+
+
 class BuildOptions:
     def __init__(self,
                  includes: Optional[List[Union[Path, str]]] = None,
                  cflags: Optional[List[str]] = None,
                  post_cflags: Optional[List[str]] = None):
-        self.includes = _list_none(includes)
+        self.includes = list(map(_unify_path, _list_none(includes)))
         self.cflags = _list_none(cflags)
         self.post_cflags = _list_none(post_cflags)
 
@@ -106,7 +123,6 @@ class BuildOptions:
         return BuildOptions(self.includes.copy(), self.cflags.copy(),
                             self.post_cflags.copy())
 
-
     def merge(self, opt: "BuildOptions"):
         res = self.copy()
         res.includes.extend(opt.includes)
@@ -121,7 +137,7 @@ class LinkOptions:
                  libpaths: Optional[List[Union[Path, str]]] = None,
                  libs: Optional[List[str]] = None,
                  ldflags: Optional[List[str]] = None):
-        self.libpaths = _list_none(libpaths)
+        self.libpaths = list(map(_unify_path, _list_none(libpaths)))
         self.libs = _list_none(libs)
         self.ldflags = _list_none(ldflags)
 
@@ -145,17 +161,20 @@ class BaseWritter(Writer):
                  compiler_link_opts: Dict[str, LinkOptions],
                  compiler_to_path: Dict[str, str],
                  linker_to_path: Dict[str, str],
+                 out_root: Optional[Union[Path, str]] = None,
                  width=78):
         # TODO check available compilers by subprocess.
         self._sstream = io.StringIO()
         super().__init__(self._sstream, width)
-
+        self.out_root = out_root
         self._build_dir = Path(build_dir).resolve()
-        self._suffix_to_cl = {}
+        self._suffix_to_compiler_var = {}
         self._suffix_to_rule = {}
         self.compiler_build_opts = compiler_build_opts
         self.compiler_link_opts = compiler_link_opts
-        self._compiler_var_to_name = {}
+        self._compiler_var_to_compiler = {}
+        self._compiler_to_compiler_var = {}
+
         self.compiler_to_path = compiler_to_path
         self.linker_to_path = linker_to_path
         suf_to_c_items = list(suffix_to_compiler.items())
@@ -170,37 +189,70 @@ class BaseWritter(Writer):
             else:
                 self.variable(compiler_name, compiler)
 
-            self._suffix_to_cl[suffix] = compiler_name
-            self._compiler_var_to_name[compiler_name] = compiler
-        self.variable("msvc_deps_prefix",  os.getenv("CCIMPORT_MSVC_DEPS_PREFIX", "Note: including file:"))
+            self._suffix_to_compiler_var[suffix] = compiler_name
+            self._compiler_var_to_compiler[compiler_name] = compiler
+            self._compiler_to_compiler_var[compiler] = compiler_name
+
+        self.variable(
+            "msvc_deps_prefix",
+            os.getenv("CCIMPORT_MSVC_DEPS_PREFIX", "Note: including file:"))
 
     @property
     def content(self) -> str:
         return self._sstream.getvalue()
 
-    def gcc_build_setup(self, name, compiler, compiler_var,
-                        opts: BuildOptions):
+    def gcc_build_setup(self,
+                        name,
+                        compiler,
+                        compiler_var,
+                        opts: BuildOptions,
+                        pch: bool = False,
+                        use_pch: bool = False):
         global_build_opts = self.compiler_build_opts.get(
             compiler, BuildOptions())
         opts = opts.merge(global_build_opts)
-        includes = " ".join(["-I \"{}\"".format(str(i)) for i in opts.includes])
-        cflags = opts.cflags
+        includes = " ".join(
+            ["-I \"{}\"".format(str(i)) for i in opts.includes])
+        cflags = opts.cflags.copy()
         post_cflags = opts.post_cflags
+        if pch:
+            # remove_flags = ["-fPIC", "-O3", "-std=c++14"]
+            # for f in remove_flags:
+            #     if f in cflags:
+            #         cflags.remove(f)
+            cflags.append("-x")
+            cflags.append("c++-header")
+
         cflags = " ".join(cflags)
         post_cflags = " ".join(post_cflags)
+
         rule_name = name + "_cxx_{}".format(compiler_var)
-        self.rule(
-            rule_name,
-            "${} -MMD -MT $out -MF $out.d {} {} -c $in -o $out {}".format(compiler_var, includes, cflags, post_cflags),
-            description="compile $out",
-            depfile="$out.d",
-            deps="gcc")
+        if pch:
+            rule_name += "_pch"
+        if use_pch:
+            rule_name += "_with_pch"
+        compile_stmt = "${} -MMD -MT $out -MF $out.d {} {} -c $in -o $out {}"
+        if use_pch:
+            compile_stmt = "${} -MMD -MT $out -MF $out.d {} {} -include $pch -c $in -o $out {}"
+        if pch:
+            compile_stmt = "${} {} {} -c $in -o $out {}"
+            self.rule(rule_name,
+                      compile_stmt.format(compiler_var, includes, cflags,
+                                          post_cflags),
+                      description="compile pch $out")
+        else:
+            self.rule(rule_name,
+                      compile_stmt.format(compiler_var, includes, cflags,
+                                          post_cflags),
+                      description="compile $out",
+                      depfile="$out.d",
+                      deps="gcc")
+
         self.newline()
         return rule_name
 
     def gcc_link_setup(self, name, linker, linker_name, opts: LinkOptions):
-        global_build_opts = self.compiler_link_opts.get(
-            linker, LinkOptions())
+        global_build_opts = self.compiler_link_opts.get(linker, LinkOptions())
         opts = opts.merge(global_build_opts)
         ldflags = " ".join(opts.ldflags)
         libs = opts.libs
@@ -210,25 +262,33 @@ class BaseWritter(Writer):
             lib_flag = "-l" + str(splits[-1])
             if len(splits) == 2:
                 prefix = splits[0]
-                if prefix == "path":
-                    lib_flag = splits[-1]
-                elif prefix == "file":
+                if prefix == "file":
                     lib_flag = "-l:" + splits[-1]
+                elif prefix == "raw":
+                    lib_flag = splits[-1]
                 else:
-                    raise NotImplementedError("unsupported lib prefix. supported: static and path")
+                    raise NotImplementedError(
+                        "unsupported lib prefix. supported: file/raw::your_flag"
+                    )
             lib_flags.append(lib_flag)
         libs_str = " ".join(lib_flags)
-        libpaths_str = " ".join(["-L \"{}\"".format(str(l)) for l in opts.libpaths])
+        libpaths_str = " ".join(
+            ["-L \"{}\"".format(str(l)) for l in opts.libpaths])
         rule_name = name + "_ld_{}".format(linker_name)
-        self.rule(
-            rule_name,
-            "${} $in {} {} {} -o $out".format(linker_name, libs_str, libpaths_str, ldflags),
-            description="link $out")
+        self.rule(rule_name,
+                  "${} $in {} {} {} -o $out".format(linker_name, libs_str,
+                                                    libpaths_str, ldflags),
+                  description="link $out")
         self.newline()
         return rule_name
 
-    def msvc_build_setup(self, name, compiler, compiler_var,
-                         opts: BuildOptions):
+    def msvc_build_setup(self,
+                         name,
+                         compiler,
+                         compiler_var,
+                         opts: BuildOptions,
+                         pch: bool = False,
+                         use_pch: bool = False):
         global_build_opts = self.compiler_build_opts.get(
             compiler, BuildOptions())
         opts = opts.merge(global_build_opts)
@@ -239,64 +299,84 @@ class BaseWritter(Writer):
         cflags = " ".join(cflags)
         post_cflags = " ".join(post_cflags)
         rule_name = name + "_cxx_{}".format(compiler_var)
-        self.rule(
-            rule_name,
-            "${} {} {} /showIncludes -c $in /Fo$out {}".format(compiler_var, includes, cflags, post_cflags),
-            deps="msvc"
-        )
+        if pch:
+            rule_name += "_pch"
+        if use_pch:
+            rule_name += "_with_pch"
+        compile_stmt = "${} {} {} /nologo /showIncludes -c $in /Fo$out {}"
+        if pch:
+            compile_stmt = "${} {} {} /nologo /showIncludes -c $in /Yc$out {}"
+        if use_pch:
+            compile_stmt = "${} {} {} /nologo /showIncludes /Yu$pch -c $in /Fo$out {}"
+            self.rule(rule_name,
+                    compile_stmt.format(
+                        compiler_var, includes, cflags, post_cflags))
+        else:
+            self.rule(rule_name,
+                    compile_stmt.format(
+                        compiler_var, includes, cflags, post_cflags),
+                    deps="msvc")
+
         self.newline()
         return rule_name
 
     def msvc_link_setup(self, name, linker, linker_name, opts: LinkOptions):
-        global_build_opts = self.compiler_link_opts.get(
-            linker, LinkOptions())
+        global_build_opts = self.compiler_link_opts.get(linker, LinkOptions())
         opts = opts.merge(global_build_opts)
         ldflags = " ".join(opts.ldflags)
         libs_str = " ".join([str(l) + ".lib" for l in opts.libs])
         libpaths_str = " ".join(
             ["/LIBPATH:\"{}\"".format(str(l)) for l in opts.libpaths])
         rule_name = name + "_ld_{}".format(linker_name)
-        self.rule(
-            rule_name,
-            "${} /link /nologo $in {} {} {} /out:$out".format(linker_name, libs_str, libpaths_str, ldflags),
-            description="link msvc $out")
+        self.rule(rule_name,
+                  "${} /link /nologo $in {} {} {} /out:$out".format(
+                      linker_name, libs_str, libpaths_str, ldflags),
+                  description="link msvc $out")
         self.newline()
         return rule_name
 
-    def nvcc_build_setup(self, name, compiler, compiler_var,
-                         opts: BuildOptions):
+    def nvcc_build_setup(self,
+                         name,
+                         compiler,
+                         compiler_var,
+                         opts: BuildOptions,
+                         pch: bool = False,
+                         use_pch: bool = False):
+        assert not pch and not use_pch, "nvcc don't support pch"
         global_build_opts = self.compiler_build_opts.get(
             compiler, BuildOptions())
         opts = opts.merge(global_build_opts)
-        includes = " ".join(["-I \"{}\"".format(str(i)) for i in opts.includes])
+        includes = " ".join(
+            ["-I \"{}\"".format(str(i)) for i in opts.includes])
         cflags = opts.cflags
         post_cflags = opts.post_cflags
         cflags = " ".join(cflags)
         post_cflags = " ".join(post_cflags)
         rule_name = name + "_cuda_{}".format(compiler_var)
         MMD = "-MD" if compat.InWindows else "-MMD"
-        self.rule(
-            rule_name,
-            "${} {} -MT $out -MF $out.d {} {} -c $in -o $out {}".format(compiler_var, MMD, includes, cflags, post_cflags),
-            description="nvcc cxx $out",
-            depfile="$out.d",
-            deps="gcc")
+        self.rule(rule_name,
+                  "${} {} -MT $out -MF $out.d {} {} -c $in -o $out {}".format(
+                      compiler_var, MMD, includes, cflags, post_cflags),
+                  description="nvcc cxx $out",
+                  depfile="$out.d",
+                  deps="gcc")
         self.newline()
         return rule_name
 
     def nvcc_link_setup(self, name, linker, linker_name, opts: LinkOptions):
-        global_build_opts = self.compiler_link_opts.get(
-            linker, LinkOptions())
+        global_build_opts = self.compiler_link_opts.get(linker, LinkOptions())
         opts = opts.merge(global_build_opts)
         ldflags = " ".join(opts.ldflags)
         libs_str = " ".join(["-l \"{}\"".format(str(l)) for l in opts.libs])
-        libpaths_str = " ".join(["-L \"{}\"".format(str(l)) for l in opts.libpaths])
+        libpaths_str = " ".join(
+            ["-L \"{}\"".format(str(l)) for l in opts.libpaths])
         rule_name = name + "_ld_{}".format(linker_name)
-        libpaths_str = " ".join(["-L \"{}\"".format(str(l)) for l in opts.libpaths])
-        self.rule(
-            rule_name,
-            "${} $in {} {} {} -o $out".format(linker_name, libs_str, libpaths_str, ldflags),
-            description="link $out")
+        libpaths_str = " ".join(
+            ["-L \"{}\"".format(str(l)) for l in opts.libpaths])
+        self.rule(rule_name,
+                  "${} $in {} {} {} -o $out".format(linker_name, libs_str,
+                                                    libpaths_str, ldflags),
+                  description="link $out")
         self.newline()
         return rule_name
 
@@ -331,20 +411,37 @@ class BaseWritter(Writer):
         else:
             raise NotImplementedError
 
-    def create_build_rule(self, compiler_name, target_name,
-                          opts: BuildOptions):
-        compiler = self._compiler_var_to_name[compiler_name]
+    def create_build_rule(self,
+                          compiler_name,
+                          target_name,
+                          opts: BuildOptions,
+                          pch: bool = False,
+                          use_pch: bool = False):
+        compiler = self._compiler_var_to_compiler[compiler_name]
         if compiler == "g++" or compiler == "clang++":
             # self.variable(compiler_name, "g++")
             return self.gcc_build_setup(target_name, compiler, compiler_name,
-                                        opts)
+                                        opts, pch, use_pch)
         elif compiler == "cl":
             # self.variable(compiler_name, "cl")
             return self.msvc_build_setup(target_name, compiler, compiler_name,
-                                         opts)
+                                         opts, pch, use_pch)
         elif compiler == "nvcc":
             # self.variable(compiler_name, "nvcc")
             return self.nvcc_build_setup(target_name, compiler, compiler_name,
+                                         opts, pch, use_pch)
+        else:
+            raise NotImplementedError
+
+    def create_pch_rule(self, compiler_name, target_name, opts: BuildOptions):
+        compiler = self._compiler_var_to_compiler[compiler_name]
+        if compiler == "g++":
+            # self.variable(compiler_name, "g++")
+            return self.gcc_build_setup(target_name, compiler, compiler_name,
+                                        opts)
+        elif compiler == "clang++":
+            # self.variable(compiler_name, "cl")
+            return self.msvc_build_setup(target_name, compiler, compiler_name,
                                          opts)
         else:
             raise NotImplementedError
@@ -356,22 +453,106 @@ class BaseWritter(Writer):
                    link_opts: LinkOptions,
                    sources: List[Union[Path, str]],
                    target_filename: str,
-                   shared=False):
-        source_paths = [Path(p) for p in sources]
+                   shared=False,
+                   pch_to_sources: Optional[Dict[Union[str, Path],
+                                                 List[Union[str,
+                                                            Path]]]] = None):
+        source_paths = [_unify_path(p) for p in sources]
+        if pch_to_sources is None:
+            pch_to_sources = {}
+        unified_pch_to_sources = {}  # type: Dict[Path, List[Path]]
+        for k, v in pch_to_sources.items():
+            k_u = _unify_path(k)
+            if k_u not in unified_pch_to_sources:
+                unified_pch_to_sources[k_u] = []
+            unified_pch_to_sources[k_u].extend(_unify_path(p) for p in v)
         path_to_rule = {}
-        compiler_to_rule = {}
+        pch_to_rule = {}
+        compiler_to_rule = {}  # type: Dict[str, str]
+        compiler_to_pch_rule = {}  # type: Dict[str, str]
+
+        pch_rule_name = ""
+        path_to_pch_obj = {}
+        path_to_pch = {}
+
         self.newline()
+        # determine PCH compiler
+        pch_compiler = ""
+        for pch_sources in unified_pch_to_sources.values():
+            pch_compiler_determined = False
+            for p in pch_sources:
+                suffix = p.suffix
+                compiler_var = self._suffix_to_compiler_var[suffix]
+                compiler = self._compiler_var_to_compiler[compiler_var]
+                if compiler in ALL_SUPPORTED_PCH_COMPILER:
+                    pch_compiler = compiler
+                    pch_compiler_determined = True
+                    break
+            if pch_compiler_determined:
+                break
+        name_pool = UniqueNamePool()
+        if pch_compiler:
+            pch_compiler_name = self._compiler_to_compiler_var[pch_compiler]
+            pch_rule_name = self.create_build_rule(
+                pch_compiler_name,
+                target_name,
+                compiler_to_option[pch_compiler],
+                pch=True)
+            for pch_path, sources_use_pch in unified_pch_to_sources.items():
+                assert pch_path.exists()
+                pch_valid_count = 0
+                valid_source_use_pch = [] # type: List[Path]
+                for source_path in sources_use_pch:
+                    assert source_path.exists()
+                    suffix = source_path.suffix
+                    compiler_var = self._suffix_to_compiler_var[suffix]
+                    compiler = self._compiler_var_to_compiler[compiler_var]
+                    if compiler == pch_compiler:
+                        pch_valid_count += 1
+                        valid_source_use_pch.append(source_path)
+                if pch_valid_count > 1:
+                    pch_obj_path = str(
+                        self._create_output_path(
+                            pch_path,
+                            name_pool,
+                            suffix=COMPILER_TO_PCH_SUFFIX[pch_compiler]))
+                    if pch_compiler in PCH_COMPILER_NEED_SOURCE:
+                        self.build(pch_obj_path, pch_rule_name, [str(pch_path)] + valid_source_use_pch)
+                    else:
+                        self.build(pch_obj_path, pch_rule_name, str(pch_path))
+                    for source_path in sources_use_pch:
+                        assert source_path.exists()
+                        suffix = source_path.suffix
+                        compiler_var = self._suffix_to_compiler_var[suffix]
+                        compiler = self._compiler_var_to_compiler[compiler_var]
+                        if compiler == pch_compiler:
+                            path_to_pch_obj[source_path] = str(pch_obj_path)
+                            path_to_pch[source_path] = str(pch_path)
+
         for p in source_paths:
             suffix = p.suffix
-            compiler_var = self._suffix_to_cl[suffix]
-            compiler = self._compiler_var_to_name[compiler_var]
-            if compiler in compiler_to_rule:
-                rule_name = compiler_to_rule[compiler]
+            compiler_var = self._suffix_to_compiler_var[suffix]
+            compiler = self._compiler_var_to_compiler[compiler_var]
+            if p in path_to_pch_obj and compiler == pch_compiler:
+                if compiler in compiler_to_pch_rule:
+                    rule_name = compiler_to_pch_rule[compiler]
+                else:
+                    compiler = self._compiler_var_to_compiler[compiler_var]
+                    rule_name = self.create_build_rule(
+                        compiler_var,
+                        target_name,
+                        compiler_to_option[compiler],
+                        use_pch=True)
+                    compiler_to_pch_rule[compiler] = rule_name
             else:
-                compiler = self._compiler_var_to_name[compiler_var]
-                rule_name = self.create_build_rule(
-                    compiler_var, target_name, compiler_to_option[compiler])
-                compiler_to_rule[compiler] = rule_name
+                if compiler in compiler_to_rule:
+                    rule_name = compiler_to_rule[compiler]
+                else:
+                    compiler = self._compiler_var_to_compiler[compiler_var]
+                    rule_name = self.create_build_rule(
+                        compiler_var, target_name,
+                        compiler_to_option[compiler])
+                    compiler_to_rule[compiler] = rule_name
             path_to_rule[p] = rule_name
         # for k, v in compiler_to_option.items():
         link_opts = link_opts.copy()
@@ -387,26 +568,52 @@ class BaseWritter(Writer):
         else:
             target_path = self._build_dir / target_filename
         obj_files = []
-        name_pool = UniqueNamePool()
         for p in source_paths:
             assert p.exists()
-            suffix = ".o"
-            file_name = name_pool(p.name)
-            obj = (self._build_dir / (file_name + suffix))
-            assert obj.parent.exists()
-            obj = str(obj)
-            obj_files.append(obj)
+            obj_path = str(self._create_output_path(p, name_pool))
+            obj_files.append(obj_path)
             rule = path_to_rule[p]
-            self.build(obj, rule, str(p))
+            if p in path_to_pch_obj:
+                pch_obj = path_to_pch_obj[p]
+                pch_path = path_to_pch[p]
+
+                self.build(obj_path,
+                           rule,
+                           str(p),
+                           variables={"pch": pch_path},
+                           implicit=[pch_obj])
+            else:
+                self.build(obj_path, rule, str(p))
+
         self.newline()
         self.build(str(target_path), link_rule, obj_files)
         self.build(target_name, "phony", str(target_path))
         self.default(target_name)
 
+    def _create_output_path(self,
+                            p: Path,
+                            name_pool: UniqueNamePool,
+                            suffix: str = ".o"):
+        source_out_parent = self._build_dir
+        if self.out_root is not None:
+            out_root = Path(self.out_root)
+            try:
+                relative = p.parent.relative_to(out_root)
+                source_out_parent = self._build_dir / relative
+            except ValueError:
+                source_out_parent = self._build_dir
+        source_out_parent.mkdir(exist_ok=True, parents=True, mode=0o755)
+        obj_path_no_suffix = (source_out_parent / (p.name))
+        obj_path_no_suffix = Path(name_pool(str(obj_path_no_suffix)))
+        obj_path = obj_path_no_suffix.parent / (obj_path_no_suffix.name +
+                                                suffix)
+        assert obj_path.parent.exists()
+        return obj_path
+
     def add_shared_target(self, target_name: str,
                           compiler_to_option: Dict[str, BuildOptions], linker,
-                          link_opts: LinkOptions,
-                          sources: List[Union[Path, str]],
+                          link_opts: LinkOptions, sources: List[Union[Path,
+                                                                      str]],
                           target_filename: str):
         return self.add_target(target_name, compiler_to_option, linker,
                                link_opts, sources, target_filename, True)
@@ -436,10 +643,7 @@ COMMON_NVCC_FLAGS = [
     '-Xcompiler=\"-fPIC\"', '-Xcompiler=\'-O3\''
 ]
 
-COMMON_NVCC_FLAGS_WINDOWS = [
-    '--expt-relaxed-constexpr', '-Xcompiler=\"/O2\"'
-]
-
+COMMON_NVCC_FLAGS_WINDOWS = ['--expt-relaxed-constexpr', '-Xcompiler=\"/O2\"']
 
 COMMON_MSVC_FLAGS = [
     '/MD', '/wd4819', '/wd4251', '/wd4244', '/wd4267', '/wd4275', '/wd4018',
@@ -501,8 +705,8 @@ def _default_target_filename(target, shared=True):
             return target
 
 
-def fill_build_flags(build_flags: Optional[Dict[str, List[str]]]
-                     ) -> Dict[str, List[str]]:
+def fill_build_flags(
+        build_flags: Optional[Dict[str, List[str]]]) -> Dict[str, List[str]]:
     """fill compile/link compiler-to-list flags with all compiler.
     """
     if build_flags is None:
@@ -513,8 +717,8 @@ def fill_build_flags(build_flags: Optional[Dict[str, List[str]]]
     return build_flags
 
 
-def fill_link_flags(link_flags: Optional[Dict[str, List[str]]]
-                    ) -> Dict[str, List[str]]:
+def fill_link_flags(
+        link_flags: Optional[Dict[str, List[str]]]) -> Dict[str, List[str]]:
     """fill compile/link compiler-to-list flags with all compiler.
     """
     if link_flags is None:
@@ -529,7 +733,7 @@ def group_dict_by_split(data: Dict[str, List[Any]], split: str = ","):
     """convert {gcc,clang++: [xxx], clang++: [yyy]}
     to {gcc: [xxx], clang++: [xxx, yyy]}
     """
-    new_data = OrderedDict() # type: Dict[str, List[Any]]
+    new_data = OrderedDict()  # type: Dict[str, List[Any]]
     for k, v in data.items():
         ks = k.split(split)
         for k_ in ks:
@@ -540,19 +744,23 @@ def group_dict_by_split(data: Dict[str, List[Any]], split: str = ","):
     return new_data
 
 
-def create_simple_ninja(target,
-                        build_dir,
-                        sources,
-                        includes=None,
-                        libs=None,
-                        libpaths=None,
-                        compile_opts=None,
-                        link_opts=None,
-                        target_filename=None,
-                        additional_cflags=None,
-                        additional_lflags=None,
-                        suffix_to_compiler=None,
-                        shared=False):
+def create_simple_ninja(
+    target,
+    build_dir,
+    sources,
+    includes=None,
+    libs=None,
+    libpaths=None,
+    compile_opts=None,
+    link_opts=None,
+    target_filename=None,
+    additional_cflags=None,
+    additional_lflags=None,
+    suffix_to_compiler=None,
+    out_root: Optional[Union[Path, str]] = None,
+    shared=False,
+    pch_to_sources: Optional[Dict[Union[str, Path],
+                                  List[Union[str, Path]]]] = None):
     default_suffix_to_compiler = _default_suffix_to_compiler()
     suffix_to_compiler_ = default_suffix_to_compiler
     if suffix_to_compiler is not None:
@@ -569,7 +777,7 @@ def create_simple_ninja(target,
             target_filename = str(path.parent /
                                   _default_target_filename(path.stem, shared))
     writer = BaseWritter(suffix_to_compiler_, build_dir, build_options,
-                         link_options, OrderedDict(), OrderedDict())
+                         link_options, OrderedDict(), OrderedDict(), out_root)
     target_build_opt = BuildOptions(includes, compile_opts)
     target_build_options = OrderedDict()
     additional_cflags = fill_build_flags(additional_cflags)
@@ -583,27 +791,32 @@ def create_simple_ninja(target,
     link_opts = LinkOptions(libpaths, libs, link_opts)
     link_opts.ldflags.extend(additional_lflags[linker])
     writer.add_target(target, target_build_options, linker, link_opts, sources,
-                      target_filename, shared)
+                      target_filename, shared, pch_to_sources)
     return writer.content, target_filename
 
-def build_simple_ninja(target,
-                       build_dir,
-                       sources,
-                       includes=None,
-                       libs=None,
-                       libpaths=None,
-                       compile_opts=None,
-                       link_opts=None,
-                       target_filename=None,
-                       additional_cflags=None,
-                       additional_lflags=None,
-                       suffix_to_compiler=None,
-                       verbose=False,
-                       shared=True):
+
+def build_simple_ninja(
+    target,
+    build_dir,
+    sources,
+    includes=None,
+    libs=None,
+    libpaths=None,
+    compile_opts=None,
+    link_opts=None,
+    target_filename=None,
+    additional_cflags=None,
+    additional_lflags=None,
+    suffix_to_compiler=None,
+    out_root: Optional[Union[Path, str]] = None,
+    verbose=False,
+    shared=True,
+    pch_to_sources: Optional[Dict[Union[str, Path],
+                                  List[Union[str, Path]]]] = None):
     ninja_content, target_filename = create_simple_ninja(
         target, build_dir, sources, includes, libs, libpaths, compile_opts,
         link_opts, target_filename, additional_cflags, additional_lflags,
-        suffix_to_compiler, shared)
+        suffix_to_compiler, out_root, shared, pch_to_sources)
     build_dir = Path(build_dir).resolve()
     with (build_dir / "build.ninja").open("w") as f:
         f.write(ninja_content)
@@ -612,11 +825,13 @@ def build_simple_ninja(target,
     if verbose:
         cmds.append("-v")
     if compat.Python3_7AndLater:
-        proc = subprocess.Popen(cmds, cwd=str(build_dir),
+        proc = subprocess.Popen(cmds,
+                                cwd=str(build_dir),
                                 stdout=subprocess.PIPE,
                                 text=True)
     else:
-        proc = subprocess.Popen(cmds, cwd=str(build_dir),
+        proc = subprocess.Popen(cmds,
+                                cwd=str(build_dir),
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True)
     output = ''
