@@ -1,9 +1,10 @@
-from pathlib import Path
 import subprocess
+from pathlib import Path
 
 import ccimport
 from ccimport.utils import tempdir
 
+
 def test_cpp_build():
     source = ccimport.autoimport([Path(__file__).parent / "source.cc"],
                                  Path(__file__).parent / "source")
@@ -11,12 +12,19 @@ def test_cpp_build():
     obj = source.TestClass(5)
     assert obj.add(3) == 8
 
+
 def test_cpp_exec_build():
     with tempdir() as tempd:
-        source = ccimport.ccimport([Path(__file__).parent / "executable.cc"],
-                                    tempd / "executable",
-                                    shared=False,
-                                    load_library=False)
-        
+        sources = [Path(__file__).parent / "executable.cc", Path(__file__).parent / "source.cc"]
+        p2s = {Path(__file__).parent / "some_pch.h": sources}
+        source = ccimport.ccimport(sources,
+                                   tempd / "executable",
+                                   shared=False,
+                                   load_library=False,
+                                   pch_to_sources=p2s)
+
         output = subprocess.check_output([str(source)])
         assert output.decode("utf-8").strip() == "hello ccimport!"
+
+if __name__ == "__main__":
+    test_cpp_exec_build()
\ No newline at end of file
